/* imidiacy - Record a MIDI file from the computer keyboard 
version 2011.4.16 Copyright (C) 2006-2011 Andy Allinger 

***********************************************************************
This program is free software, you can redistribute it and/or modify it 
under the terms of the GNU General Public License, as published by the 
Free Software Foundation, version 2.

This program is distributed in hope that it will be useful but WITHOUT 
ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not write the Free Software Foundation, Inc. 
59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

This program produces standard MIDI files based on presses of the PC
keyboard.  The events Note Off, Note On, Aftertouch, and tempo change
are supported.  Output is to the monitor and soundcard or PC speaker.

For the most recent released version, visit:
imidiacy.sourceforge.net

Or, contact the author at:
andy_a@users.sourceforge.net

***********************************************************************/

/* Turn a variable-byte sequence into an integer */

unsigned long read_var_bytes(unsigned char *midi_fmt)
{
  unsigned long number = 0 ;
  number = (midi_fmt[0] & 127) ;
  if (! (midi_fmt[0] & 128)) return number ;

  number <<= 7 ; number |= (midi_fmt[1] & 127) ;
  if (! (midi_fmt[1] & 128)) return number ;

  number <<= 7; number |= (midi_fmt[2] & 127) ;
  if (! (midi_fmt[2] & 128)) return number ;

  number <<= 7; number |= (midi_fmt[3] & 127) ;

  return number ;

}  /* End Function read_var_bytes */

/**********************************************************************/

/*
Convert delta-time, in ticks, from a long integer to the variable-byte 
format used in the midi file.  Return the number of bytes, putting 
output into a char array
*/

int write_var_bytes(unsigned long number, unsigned char *midi_fmt)
{
  int bytes_rqd ;

/* The "loop unrolled" version of the function */
  if (number <= 127) {
    bytes_rqd = 1;
    midi_fmt[0] = number & 127 ;
    return bytes_rqd ;
  }
  else if (number <= 16383) {
    bytes_rqd = 2 ;
    midi_fmt[0] = ((number & 16256) >>7) | 128 ; 
    midi_fmt[1] = (number & 127) ;
    return bytes_rqd;
  }
  else if (number <= 2097151) {
    bytes_rqd = 3 ;
    midi_fmt[0] = ((number & 2080768) >>14) | 128 ;
    midi_fmt[1] = ((number & 16256) >>7) | 128 ;
    midi_fmt[2] = (number & 127) ;
    return bytes_rqd ;
  }
  else {
    bytes_rqd = 4 ;
    midi_fmt[0] = ((number & 266338304) >>21) | 128 ;
    midi_fmt[1] = ((number & 2080768) >>14) | 128 ;
    midi_fmt[2] = ((number & 16256) >>7) | 128 ;
    midi_fmt[3] = (number & 127) ;
    return bytes_rqd ;
  }

}  /* End Function write_var_bytes */

/**********************************************************************/
/*
This function just counts the bytes, without writing to the character array
*/

int count_var_bytes(unsigned long number)
{
  int bytes_rqd ;

  bytes_rqd = 1 ;
  if (number > 127) bytes_rqd = 2 ;
  if (number > 16383) bytes_rqd = 3 ;
  if (number > 2097151) bytes_rqd = 4 ;

  return bytes_rqd ;

} /* End Function count_var_bytes */

/**********************************************************************/
