***** ******************************************************************
       PROGRAM mixmidi  ! music analysis and synthesis
***** ******************************************************************

       IMPLICIT NONE
       LOGICAL fexist, x
       INTEGER argcount, numerr, filesize, i, j, L, n, q
       INTEGER num_q ; PARAMETER (num_q = 25)
       INTEGER ontime, offtime, pitch
       CHARACTER * 256 indev
       CHARACTER * 256 text
       
       REAL change ; DIMENSION change(0:12*num_q-1, 0:12*num_q-1, 0:1)
       REAL pip, tend ; PARAMETER (pip = 1.0)
 
***** ******************************************************************
       CALL SRAND(TIME())
! interpret command line
       argcount = IARGC()
       x = .FALSE.

!           permit for testing
!       IF (argcount < 1) CALL usage_bail()
      
       CALL GETARG(1, indev)
       IF (indev(1:1) .EQ. '-') THEN
         IF (indev(2:2) .EQ. 'h' .OR. indev(2:2) .EQ. 'H') THEN
           PRINT *, 'chord - chordal analysis by the Pardo-Birmingham'
     &       // ' algorithm.'
           PRINT *,'This program is free software according to the '
     &       // 'terms of the Allegro Giftware License.'
           PRINT *, 'Measure chord transition probabiltites from a '
     &       // ' collection of notelist files:  '
           PRINT '(1X A)', ' chord -X <file1.notes> [file2.notes] ...'
           PRINT *, 'Analyze a notelist file: '
           PRINT '(1X A)', ' chord file.notes'
           STOP
         ELSE IF (indev(2:2) .EQ. 'x' .OR. indev(2:2) .EQ. 'X') THEN
           x = .TRUE.
         ELSE
           CALL usage_bail()
         END IF 
       END IF ! hyphen
       
*-----------------------------------------------------------------------
*        Code for probability extraction starts line 50
*-----------------------------------------------------------------------
       IF (x) GO TO 50       
       
            !           default notefile for testing
       IF (argcount .NE. 1) indev = 'susannah.q.notes'
!       IF (argcount .NE. 1) CALL usage_bail()
        
*          load configuration data
        OPEN (UNIT=3, FILE='chord.cfg', STATUS='OLD', FORM='UNFORMATTED'
     &     , ACCESS='SEQUENTIAL', ERR=90)
        READ (UNIT=3, ERR=90) change
        CLOSE (UNIT=3, ERR=90)
        
*          open notelist file        
        INQUIRE (FILE=indev, EXIST=fexist, IOSTAT=numerr, ERR=90)
        IF (.NOT. fexist) STOP 'File not found'
        IF (numerr .NE. 0) STOP 'Error opening input file'
        OPEN (UNIT=1, FILE = indev, STATUS = 'OLD', FORM = 'FORMATTED')
        REWIND (UNIT=1)

! Find length of file
        filesize = 0
        DO
          READ (UNIT=1, FMT=*, END=20, ERR=90) 
     &                                  text, ontime, offtime, pitch
          filesize = filesize + 1
        END DO
!  23    FORMAT (A)
  20    REWIND (UNIT=1)

*         time of last event
        tend = FLOAT(offtime) / 1000
        n = tend / pip + 1
        q = (n-1) / 2 + 1
        IF (n < 7) q = 1           ! don't overfit just a few pips         
        
        CALL analyze(change, filesize*2, n, q, pip, tend, x)
        CLOSE (UNIT = 1, ERR = 90)
        GO TO 99
      
*-----------------------------------------------------------------------      
  50   CONTINUE      !  extract configuration data from multiple files
*                       initialize change to zero
*-----------------------------------------------------------------------
       DO L = 0, 1
         DO j = 0, 12 * num_q - 1
           DO i = 0, 12 * num_q - 1
             change(i, j, L) = 0.0
           END DO
         END DO
       END DO

*         check for existence of input files
       DO j = 2, argcount
         CALL GETARG(j, indev)
         INQUIRE (FILE=indev, EXIST=fexist, IOSTAT=numerr, ERR=90)
         IF (.NOT. fexist) STOP 'File not found'
       END DO
       
*          open input files
       DO j = 2, argcount
         CALL GETARG(j, indev)
                
         OPEN (UNIT=1, FILE = indev, STATUS = 'OLD', FORM = 'FORMATTED',
     &    IOSTAT=numerr)
         IF (numerr .NE. 0) STOP 'Trouble opening input file'
         REWIND (UNIT=1)
         filesize = 0
         DO
           READ (UNIT=1, FMT=23, END=30, ERR=90) text
           filesize = filesize + 1
         END DO
  23     FORMAT (A)
  30     REWIND (UNIT=1)
         
         CALL analyze(change, filesize*2, n, q, pip, tend, x)
         CLOSE (UNIT = 1, ERR = 90)
       END DO

*-----------------------------------------------------------------------
*       save log probabilities of collection
*-----------------------------------------------------------------------
       CALL ctlp(change)
       INQUIRE (FILE='chord.cfg', EXIST=fexist, IOSTAT=numerr)
       IF (numerr .NE. 0) STOP 'Trouble inquiring of chord.cfg'
       IF (fexist) CALL UNLINK('chord.cfg', numerr)
       IF (numerr .NE. 0) STOP 'Trouble deleting file'
       
       OPEN (UNIT=3, FILE='chord.cfg', STATUS='NEW', FORM='UNFORMATTED'
     &     , ACCESS='SEQUENTIAL', ERR=90)
       REWIND (UNIT=3)
       WRITE (UNIT=3, ERR=90) change
       CLOSE (UNIT=3, ERR=90)
                   
       GO TO 99
      
***** ******************************************************************
*     termination block
  90   STOP 'file I/O error'
  99   END                   ! of mixmidi



*-----------------------------------------------------------------------
*       Read in file, call chord subroutine
*-----------------------------------------------------------------------
       SUBROUTINE analyze(change, n, npips, npts, pip, tend, x)
       
       IMPLICIT NONE
       INTEGER num_q ; PARAMETER (num_q = 25)
       INTEGER n, npips, npts
       REAL change ; DIMENSION change(0:12*num_q-1, 0:12*num_q-1, 0:1)
       LOGICAL x

*           key-changes ("modulations" in music theory) not implemented 
       INTEGER k ; PARAMETER (k = 1)     
       INTEGER num_strata, LT
       PARAMETER (num_strata = 5,          ! fewer strata than oscillators
     &            LT = 3 )                 ! index of the tactus level

       INTEGER a, h, i, j, ontime, offtime, notenum, pitch, ind, offon,
     & kp, sharps, majmin, c, q, r, m, fault
       REAL note_time, key_time, partit, p, dur, T, S, eps, 
     &  beat, pip, tend, tprev, tnext, prob, machine, tau, snap
       DIMENSION note_time(n), pitch(n), offon(n), ind(n), dur(n/2),
     & sharps(k), majmin(k), key_time(k), q(n), r(n), kp(n/2), 
     & partit(n), T(n/2), S(n/2), beat(n,num_strata)
     
       CHARACTER*2 keysig_name(-7:7, 0:1)
       CHARACTER*3 tpc, spell(0:11, -7:7)   ! spellings for different keys
       CHARACTER*5 word
       CHARACTER*5 quality_name(0:num_q-1)  ! chord symbols
       
*             parameters for rhythm detection
       REAL vel_sal, dur_sal, har_sal
     
*           phase trajectories at each stratum     
       REAL CX(npts+1), CY(npts,num_strata), 
     &  CA(npts,num_strata), CB(npts,num_strata), CC(npts,num_strata)
     
*             These saliences are mere guesses     
       PARAMETER (vel_sal = 0.002,        ! rhythmic salience per velocity
     &            dur_sal = 0.5,          ! rhythmic salience of duration
     &            har_sal = 1.0)          ! rhythmic salience per chord change
       
*          major and minor key-signature names
       DATA keysig_name / 'Cb', 'Gb', 'Db', 'Ab', 'Eb', 'Bb', 'F',
     &  'C', 'G', 'D', 'A', 'E', 'B', 'F#', 'C#',
     &  'Ab', 'Eb', 'Bb', 'F', 'C', 'G', 'D', 'A', 'E', 'B', 'F#', 'C#',
     &  'G#', 'D#', 'A#' /
     
*-----------------------------------------------------------------------
*  tonal pitch class spellings, based on minimum distance on the line of fifths
*-----------------------------------------------------------------------
       DATA spell /                
     &  'Dbb','Db','Ebb','Eb','Fb','F','Gb','Abb','Ab','Bbb','Bb','Cb',    ! Cb
     &  'C','Db','Ebb','Eb','Fb','F','Gb','Abb','Ab','Bbb','Bb','Cb',      ! Gb
     &  'C','Db','Ebb','Eb','Fb','F','Gb','G','Ab','Bbb','Bb','Cb',        ! Db
     &  'C','Db','D','Eb','Fb','F','Gb','G','Ab','Bbb','Bb','Cb',          ! Ab
     &  'C','Db','D','Eb','Fb','F','Gb','G','Ab','A','Bb','Cb',            ! Eb
     &  'C','Db','D','Eb','E','F','Gb','G','Ab','A','Bb','Cb',             ! Bb
     &  'C','Db','D','Eb','E','F','Gb','G','Ab','A','Bb','B',              ! F
     &  'C','Db','D','Eb','E','F','F#','G','Ab','A','Bb','B',              ! C
     &  'C','C#','D','Eb','E','F','F#','G','Ab','A','Bb','B',              ! G
     &  'C','C#','D','Eb','E','F','F#','G','G#','A','Bb','B',              ! D
     &  'C','C#','D','D#','E','F','F#','G','G#','A','Bb','B',              ! A
     &  'C','C#','D','D#','E','F','F#','G','G#','A','A#','B',              ! E
     &  'C','C#','D','D#','E','E#','F#','G','G#','A','A#','B',             ! B
     &  'B#','C#','D','D#','E','E#','F#','G','G#','A','A#','B',            ! F#
     &  'B#','C#','D','D#','E','E#','F#','F##','G#','A','A#','B' /         ! C#
     
*-----------------------------------------------------------------------
*   chord quality names - Make sure this list matches template[] in chordal
*-----------------------------------------------------------------------
       DATA quality_name / 
     &  ' ', 'm', '7', 'maj7', 'm7', '6', 'm6',  
     &  'dim', 'aug', 'sus4', '-5', 
     &  'dim7', 'aug7', '7sus4', '7-5', 
     &  'add9', '9', 'maj9', 'm9', 
     &  '11', '13', '69', 'm69', '7b9', '7#9' /
     
       DATA tau / 0.5 /        ! constant for saturation of durational accent

*-----------------------------------------------------------------------
*   begin
*-----------------------------------------------------------------------
       eps = 2 * machine(1)
       
*        read a note list into two different data structures
*          the real program should be more efficient than this
       DO i = 1, n / 2
         READ (UNIT=1, FMT=*, ERR=91) word, ontime, offtime, notenum
        
        note_time(2*i-1) = FLOAT(ontime) / 1000          ! NoteOns
        pitch(2*i-1) = notenum
        offon(2*i-1) = 1
        
        note_time(2*i) = FLOAT(offtime) / 1000           ! NoteOffs
        pitch(2*i) = notenum
        offon(2*i) = 0

        kp(i) = notenum                    ! different arrays for key-finding
        dur(i) = note_time(2*i) - note_time(2*i-1)
        T(i) = FLOAT(ontime) / 1000
        
       END DO        ! next input record       
 
***** ******************************************************************        
*          sort the note times, pitches, and on/off bytes       
***** ******************************************************************        
       CALL QSORTR(note_time, ind, n)        ! sort off / on data
       CALL RORDER(note_time, ind, n)
       CALL IORDER(pitch, ind, n)
       CALL IORDER(offon, ind, n)

       CALL QSORTR(T, ind, n/2)       ! sort ontime / duration data
       CALL RORDER(T, ind, n/2)
       CALL RORDER(dur, ind, n/2) 
       
!       dump the input
        DO  i = 1, n
          PRINT 27, i, note_time(i), pitch(i)
          IF (offon(i) .EQ. 0) THEN 
            PRINT 28
          ELSE
            PRINT 29
          END IF
        END DO
  27    FORMAT ('Event Num.', I5, ' Time=', F10.5, ' Note Num.', I3, $)
  28    FORMAT (' Off')
  29    FORMAT (' On')
 
*-----------------------------------------------------------------------
*          find the key
*-----------------------------------------------------------------------
        h = 1
        CALL skey(kp, dur, n/2, sharps(h), majmin(h), fault)
        key_time(h) = 0.0 

*-----------------------------------------------------------------------
*           do the chordal analysis
*-----------------------------------------------------------------------
       CALL chordal(note_time, pitch, offon, n, key_time, sharps, 
     &  majmin, k, c, partit, q, r, p, change, x, fault)
       IF (fault .NE. 0) PRINT *, 'chordal:  error #', fault
       
*-----------------------------------------------------------------------
*   find rhythmic saliences.  
* scale duration in the manner of Toivaianen & Eerola's Midi Toolbox, citing:
*  R. Parncutt
*  "A Perceptual Model of Pulse Salience and Metrical Accent in Musical Rhythms"
*  Music Perception 11(4), 409-464
*-----------------------------------------------------------------------
       DO i = 1, n/2
         S(i) = dur_sal * (1 - SNAP(-dur(i) / tau))
       END DO
       
*         merge identical onset times and their saliences
       tprev = T(1) ; m = 1
       DO i = 2, n/2
         tnext = T(i)
         IF (ABS(tnext-tprev) < eps) THEN
           S(m) = S(m) + S(i)
         ELSE
           m = m + 1
           T(m) = tnext
           S(m) = S(i)
           tprev = tnext
         END IF
       END DO
       
*        add in harmonic salience
       a = 1
       PRINT *, 'T(1)=', T(1), ' partit(1)=', partit(1)
       DO i = 1, m
         IF (ABS(partit(a) - T(i)) < eps) THEN
           PRINT *, 'adding har_sal partit=', partit(a), ' T=', T(i)
           S(i) = S(i) + har_sal
           a = a + 1
         ELSE IF (partit(a) < T(i) .AND. a .LE. c) THEN
           a = a + 1
           PRINT *, 'analyze:  bad increment of chord change index'           
         END IF
       END DO
 
*          call the rhythmic strata routine
       CALL strata(T, S, num_strata, LT, m, npips, npts, pip,
     &  prob, CX, CY, CA, CB, CC, fault)
       IF (fault .NE. 0) PRINT *, 'strata: error #', fault

       DO i = 1, num_strata                     ! get phases
         CALL SCOMP(CX, CY(1,i), CA(1,i), CB(1,i), CC(1,i), npts,
     &       note_time, beat(1,i), n, fault)
         IF (fault .NE. 0) PRINT *, 'scomp:  error #', fault
       END DO

*          print results of analysis       
       IF (.NOT. x) THEN
         PRINT *, 'Results:  '
         PRINT 31, 'Key of ', keysig_name(sharps(h), majmin(h))
  31   FORMAT (1X, A, A2, ' ', $)
         IF (majmin(h) .EQ. 0) THEN
           PRINT *, 'Major'
         ELSE IF (majmin(h) .EQ. 1) THEN
           PRINT *, 'Minor'
         ELSE 
           PRINT *, 'bad major/minor value!'
         END IF
         
*          output for all note events
         j = 0
         DO i = 1, n
           PRINT 27, i, note_time(i), pitch(i)
           IF (offon(i) .EQ. 0) THEN 
             PRINT 32
           ELSE
             PRINT 33
           END IF
                                            
           IF (j < c) THEN
             IF (partit(j+1) .LE. note_time(i)) THEN
               j = j + 1

* Use tonal pitch class spelling of root, given a key signature
               tpc = spell(r(j), sharps(h))
               PRINT 34, tpc(1:LEN_TRIM(tpc)), quality_name(q(j))
             ELSE
               PRINT 35
             END IF
           ELSE
             PRINT 35
           END IF
           PRINT 40, beat(i,1),beat(i,2),beat(i,3),beat(i,4),beat(i,5)
         END DO
         
  32     FORMAT (' Off', $)
  33     FORMAT (' On', $)
  34     FORMAT (' Chord:  ', A, A5)
  35     FORMAT ()
  40     FORMAT ('Phase: ', 5F10.2)
       END IF !  (NOT x)

       PRINT *, 'overall harmonic probability:  ', p
       PRINT *, 'overall rhythmic probability: ', prob
       RETURN      

 91    STOP 'file i/o error'
      END  ! of analyze

************************************************************************
       SUBROUTINE usage_bail()       
         PRINT *, 'usage: chord [-X] <file1.notes> [file2.notes] [...]'
         STOP
       END ! of usage_bail
