*==== ==================================================================

* test clustering
* by Andy, 2005-2011, P.D.
      PROGRAM tkm
      
      IMPLICIT NONE      
      INTEGER mlim, M,N,K,ITER,FAULT             ! points, dimension, errors
      INTEGER i, j, kmin, kmax, o, ferr, L, r
      INTEGER iopt                         ! choose an initialization method

      REAL A(150,4), C(3, 4), iswap, Asub(150,4), T(4), y
      REAL u, ucalc, w, total, best, worst, f
      INTEGER Q(150), P(3), ind(150), sample

      LOGICAL fexist, robust, b, dump
      CHARACTER*19 fname

      DATA M,N,K,ITER,FAULT,total,worst /150,4,3,1000, 0, 0., 0./

      CALL SRAND(TIME())
      robust = .TRUE.
      f = 7
      y = 6.0
      mlim = m
      kmin = 3
      kmax = 3
      best = 1E30
      b = .true.
      dump = .false.

*        get input file
      fname = 'iris.data'
      INQUIRE (FILE = fname, EXIST = fexist, IOSTAT = ferr, ERR = 901)
      IF (.NOT. fexist) STOP 'File not found'
      IF (ferr .NE. 0) STOP 'Error opening input file'
      OPEN (UNIT = 1, FILE = fname, STATUS = 'OLD', 
     &        ACCESS = 'SEQUENTIAL', ERR=901)

***** ******************************************************************
*       start reading in data
       DO i = 1, M
        READ (UNIT = 1, FMT = 301, END = 310)A(i,1),A(i,2),A(i,3),A(i,4)
 301    FORMAT (F3.1, ',', F3.1, ',', F3.1, ',', F3.1)
       END DO
 310  CLOSE (UNIT = 1, ERR=901)
!      PRINT *, A
!      PAUSE

!         disable modulo distance
      DO j = 1, N ; T(j) = 0 ; END DO

!      compare initialization methods 
      iopt = 2

      DO o = 1, 100

*---- ------------------------------------------------------------------
*       random cluster centers (good ways to initialize are hard)
      IF (iopt .EQ. 1) THEN
         DO L = 1, K
          C(L, 1) = RAND(0)*2 + 5.
          C(L, 2) = RAND(0)*1.5 + 2.
          C(L, 3) = RAND(0)*2 + 5.
          C(L, 4) = RAND(0)*1.5 + 1.
         END DO

*       random points as centers
       ELSE IF (iopt .EQ. 2) THEN
        DO L = 1, k
        r = INT(RAND(0)*m) + 1
         DO j = 1, n
            C(L,j) = A(r,j) 
         END DO
       END DO
 
!       random cluster memberships
      ELSE IF (iopt .EQ. 3 .OR. iopt .EQ. 5) THEN
       DO L = 1, k ; P(L) = 0 ; END DO
       DO i = 1, m
        Q(i) = INT(RAND(0)*K) + 1
        P(Q(i)) = P(Q(i)) + 1
       END DO

!        hierarchical
      ELSE IF (iopt .EQ. 4) THEN
       CALL hclust(A,mlim,m,n,k,Q,P,T,y,fault)
      END IF

*        let clusters be the average of their points
      IF (iopt .EQ. 3 .OR. IOPT .EQ. 4 .OR. IOPT .EQ. 5) THEN
       DO j = 1, N
        DO L = 1,K
         C(L,j) = 0.0
        END DO
       END DO
       
       DO j = 1, n
        DO i = 1, m
          r = q(i)
          C(r,j) = C(r,j) + A(i,j)
         END DO
        END DO
       
       DO j = 1, n
        DO L = 1, k
         C(L,j) = C(L,j) / P(L)
        END DO
       END DO
      END IF
              
!       multi-stage with subsampling
      sample = m/2

      IF (iopt .EQ. 5) THEN
        DO i = 1, m ; ind(i) = i ; END DO ! make a random permutation vector
        DO i = m, m-sample+1, -1   ! partial Durstenfeld shuffle
          r = INT(RAND(0)*i) + 1   
          iswap = ind(r) ; ind(r) = ind(i) ; ind(i) = iswap
        END DO
*        print *, ind
*        pause
        
        DO j = 1, n                     ! transfer data to temp. array
          DO i = 1, sample     
            Asub(i,j) = A(ind(m-i+1),j)
          END DO
        END DO
        
        CALL kmeans(Asub, m,sample, n, C,kmax,K, Q, P, T, iter, U,fault)
      END IF

*vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
*      everything prior leads up to the next line
*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
!        PRINT *, 'Calling sub, m=', M, ' n=', n, ' k=', k
!        CALL kmeans(A,mlim,M,N,C,kmax,K,Q,P,T,ITER,u,FAULT)
!         CALL kmedians(A,mlim,M,N,C,kmax,K,Q,P,T,iter,U,fault)
         r = 1 ; y = -.75
         CALL kasc(A,M,N,R,C,kmin,k,kmax,Q,P,T,y,f,b,robust,u,fault)
!        CALL kmns(A,MLIM,M,N,C,KMAX,K,Q,P,iter,U,fault)
!        CALL kmeans_MacQueen(A,mlim,m,n,C,kmax,k,Q,P,iter,U,fault)
        IF (fault>0) PRINT *, 'Error #', fault
        IF (k .NE. 3) PRINT *, 'Made ', k, ' clusters'
        
*       residual error
      ucalc = 0.0
      DO i = 1, m
        w = 0.0
        r = q(i)
        DO j = 1, n
          w = w + (C(r,j) - A(i,j)) **2
        END DO
        ucalc = ucalc + SQRT(w)
      END DO

!      dump the output
      IF (dump .AND. ucalc < 97.3) THEN
       dump = .FALSE.
       OPEN (UNIT=2, ACCESS='SEQUENTIAL', FILE='iris.out', 
     &   STATUS='new', ERR=901)
       DO i = 1, M
         PRINT 52, Q(i)
         WRITE (UNIT=2, FMT=52) Q(i)
         DO j = 1, N        
           PRINT 51, A(i, j)
           WRITE (UNIT=2, FMT=51) A(i,j)
         END DO
        PRINT 53
         WRITE (UNIT=2, FMT=53)
       END DO
 51    FORMAT (F4.1, $)
 52    FORMAT (I2, $)
 53    FORMAT ()

       PRINT 53
       WRITE(UNIT=2, FMT=53)
       DO i = 1, K
         PRINT 52, i
         WRITE (UNIT=2, FMT=52) i
         DO j = 1, N
           PRINT 51, C(i,j)
           WRITE (unit=2, fmt=51) C(i,j)
         END DO
         PRINT 53
         WRITE (UNIT=2, FMT=53)
       END DO
       CLOSE (UNIT=2)
      END IF

      IF (MOD(o, 10) .EQ. 0) THEN        
        PRINT *, 'On try ', o, ' Distance was:  ', u, ' Calc U=', ucalc
      END IF
      
      best = MIN(best, ucalc)
      worst = MAX(worst, ucalc)
      total = total + ucalc
!      worst = MAX(worst, u)
!      total = total + u
      END DO ! next o
 
      PRINT *,'Worst case: ',worst,' best case: ',best,' Total: ',total

      GO TO 99
      
 901  STOP 'File i/o error'
 99   END 

