!     ------------------------------------------------------------------  
! var_byte.f - bit flipping routines for the MIDI variable-byte format
! (c) Andy Allinger, 2006-2011

***********************************************************************
! This program is free software, you can redistribute it and/or modify it 
! under the terms of the GNU General Public License, as published by the 
! Free Software Foundation, version 2.

! This program is distributed in hope that it will be useful but WITHOUT 
! ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
! for more details.

! You should have received a copy of the GNU General Public License along 
! with this program; if not write the Free Software Foundation, Inc. 59 
! Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

! For the most recent released version, visit:
! imidiacy.sourceforge.net

! Or, contact the author at:
! andy_a@users.sourceforge.net

***********************************************************************/

!     Convert delta-time from an integer to the variable-byte format 
!     Return value is the number of bytes, the output is passed back in 
!     the same variable.
**********************************************************************

!     ------------------------------------------------------------------
!      Count the bytes req'd to represent an integer
!     ------------------------------------------------------------------  
       INTEGER FUNCTION count_var_bytes(n)

       IMPLICIT NONE
       INTEGER n, bytes_reqd

       bytes_reqd = 1
       IF (n > 127) bytes_reqd = 2
       IF (n > 16383) bytes_reqd = 3
       IF (n > 2097151) bytes_reqd = 4
       
       count_var_bytes = bytes_reqd
       RETURN
       END

!     ------------------------------------------------------------------  
!     Convert an integer to MIDI variable-byte format
!     a new routine without loops for faster performance
!     ------------------------------------------------------------------
      INTEGER FUNCTION write_var_bytes(number, var_bytes)
      
      IMPLICIT NONE
      INTEGER number, bytes_reqd
      INTEGER*1 var_bytes ; DIMENSION var_bytes(4)

      IF (number .LE. 127) THEN
       bytes_reqd = 1
       var_bytes(1) = IAND(number, 127)
       write_var_bytes = bytes_reqd 
       RETURN

      ELSE IF (number .LE. 16383) THEN
       bytes_reqd = 2 
       var_bytes(1) = IBSET(ISHFT(IAND(number, 16256), -7), 7)
       var_bytes(2) = IAND(number, 127)
       write_var_bytes = bytes_reqd 
       RETURN
      
      ELSE IF (number .LE. 2097151) THEN
       bytes_reqd = 3 
       var_bytes(1) = IBSET(ISHFT(IAND(number, 2080768), -14), 7)
       var_bytes(2) = IBSET(ISHFT(IAND(number, 16256), -7), 7)
       var_bytes(3) = IAND(number, 127)
       write_var_bytes = bytes_reqd
       RETURN
       
      ELSE
       bytes_reqd = 4
       var_bytes(1) = IBSET(ISHFT(IAND(number, 266338304), -21), 7)
       var_bytes(2) = IBSET(ISHFT(IAND(number, 2080768), -14), 7)
       var_bytes(3) = IBSET(ISHFT(IAND(number, 16256), -7), 7)
       var_bytes(4) = IAND(number, 127)
       write_var_bytes =  bytes_reqd
       RETURN
      END IF
      
      END

!     ------------------------------------------------------------------
!      Turn the variable-byte sequence back into an integer 
!     ------------------------------------------------------------------
       INTEGER FUNCTION read_var_bytes(var_bytes)
       
       IMPLICIT NONE
       INTEGER*1 var_bytes ; DIMENSION var_bytes(4)
       INTEGER number
      
!      1 byte size
       number = IAND(var_bytes(1), 127)
       IF (.NOT. BTEST(var_bytes(1), 7)) THEN
        read_var_bytes = number
        RETURN
       END IF

!      2 byte size
       number = ISHFT(number, 7)
       number = IOR(number, IAND(var_bytes(2), 127))
       IF (.NOT. BTEST(var_bytes(2), 7)) THEN
        read_var_bytes = number
        RETURN
       END IF
       
!      3 byte size
       number = ISHFT(number, 7)
       number = IOR(number, IAND(var_bytes(3), 127))
       IF (.NOT. BTEST(var_bytes(3), 7)) THEN
        read_var_bytes = number
        RETURN
       END IF

!      4 byte size       
       number = ISHFT(number, 7)
       number = IOR(number, IAND(var_bytes(4), 127))
       read_var_bytes = number

       END

!     ------------------------------------------------------------------
!      Main program for testing purposes only
!     ------------------------------------------------------------------
      PROGRAM midi_check
      
      IMPLICIT NONE
      INTEGER i, n, big, br
      INTEGER*1 vb ; DIMENSION vb(4)
      INTEGER read_var_bytes, write_var_bytes
      
!     Loop to 2^28 -1 to test the full range of integers

      big = (2 **10) -1
      DO i = 1, big, 78
       br = write_var_bytes(i, vb)
       n = read_var_bytes(vb)
       IF ( (br < 1) .OR. (br > 4)) PRINT 800, br
       IF (i .NE. n) PRINT 801, i, n
       PRINT 802, i, vb 
      END DO
      
 800  FORMAT ('Number of bytes: ', I9, ' is out of range.')
 801  FORMAT ('The retrieved value of ', I9, ' does not match the origin
     &al value of ', I9)
 802  FORMAT ('Test value of ', I10, ' yields ', 4Z2)
     
      END
        
