/* imidiacy - Record a MIDI file from the computer keyboard 
version 2007.9.06 Copyright (C) 2006-2007 Andrew Allinger 

***********************************************************************
This program is free software, you can redistribute it and/or modify it 
under the terms of the GNU General Public License, as published by the 
Free Software Foundation, version 2.

This program is distributed in hope that it will be useful but WITHOUT 
ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not write the Free Software Foundation, Inc. 59 
Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

This program produces standard MIDI files based on presses of the PC
keyboard.  The events Note Off, Note On, Aftertouch, and tempo change
are supported.  Output is to the monitor and soundcard or PC speaker.
 
For the most recent released version, visit:
imidiacy.sourceforge.net

Or, contact the author at:
andy_a@users.sourceforge.net

***********************************************************************/

/* defines which may need to be edited */
char version_string[] = " imidiacy version 2007.9.06 (GPL)" ;
#define LINUX_VERSION

/* Standard C */
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>

/* Linux-specific includes */
#ifdef LINUX_VERSION
  #include <unistd.h>
  #include <sys/stat.h>
#endif

/* Some structures and constants. */
#include "midi_types.h"   

/* Global variables */
DBL_NODE *ptr_head ;   /* the beginning of the linked list */

/*********************************************************************/

/* The main() function queries the user, calls record, and writes the file. */

int main()
{

  /* variables for input before recording */
  int assign = 0 ;
  bool keep =          TRUE,
       use_sound =  FALSE,
       harmony = FALSE ;
  long tempo ;
  int key_sig, time_sig, bpm, velocity, channel, patch, take = 0 ;
  float legato ;
  long count_bytes ;      /* return value of record() */

/* Structure in which to retrieve user input.  It contains no pointers, 
and thus needs no initialization */
  struct S_V setting_vars ;     

/* prepare for bailing out of the program, turning off the speaker */
  signal(SIGINT, handle_signal) ;
  atexit(end_sound) ;
  atexit(hush_note) ;
  atexit(erase_list) ;
  atexit(cleanup_screen) ;

  ready_screen() ;      /* routines to set up the multimedia */
  legal(version_string) ;

/* Top of the loop.  Lines in main() above here execute only once.  Ask 
the user for input, then take it out of the structure */

Settings:
  channel = query_channel() ;
  setting_vars.channel = channel ;
  query_settings(&setting_vars) ;
  bpm = setting_vars.bpm ;
  tempo = MICROSECONDS_PER_MINUTE / bpm ;
  velocity = setting_vars.velocity ;
  legato = setting_vars.legato ;
  channel = setting_vars.channel ;
  key_sig = setting_vars.key_sig ;
  time_sig = setting_vars.time_sig ;
  harmony = setting_vars.harmony ;
  patch = setting_vars.patch ;

  use_sound = (init_sound(patch, channel) < 0) ? FALSE : TRUE ;  /* set up sound drivers */

Take:
  press_any_key() ;         /* Pause for a moment */
  ptr_head = NULL ;        /* clear the head pointer */
  draw_display(bpm, velocity, legato, version_string) ;
  count_bytes = record(tempo, velocity, legato, channel, key_sig, time_sig, use_sound, harmony) ;
       
/* if user likes the take, then write the file */
  keep = query_keep() ;  
  if (keep) {        
    ++take ;
    write_file(tempo, channel, key_sig, time_sig, patch, take, count_bytes) ;
  }                 
  erase_list() ;

/* Changing the execution to use an Assigned Go To, for clarity (and to 
heck with pedantry!) */
  assign = query_another() ;
  if (assign == 1) goto Take ;
  if (assign == 2) goto Settings ;

  return EXIT_SUCCESS ;
}

#ifdef ALLEGRO_H
  END_OF_MAIN() 
#endif

/* End Program */

/**********************************************************************/
/* the tedious work of putting all the data on disk.  Fairly straighforward. */
int write_file(long tempo, 
               int channel, 
               int key_sig, 
               int time_sig, 
               int patch, 
               int take, 
               long count_bytes)
{
/* variables to input after the record, before writing file */
  int i, c, bytes_rqd ;
  DBL_NODE *ptr_active ;
  FILE *output ;
  unsigned char var_bytes[4] ;
  char text[STRING_SIZE], 
       copyright[STRING_SIZE], 
       title[STRING_SIZE], 
       instrument[STRING_SIZE],
       filename[STRING_SIZE],
       take_str[6] ;           /* just long enough to hold "65535" */
  bool use_text, use_copyright ; 
  int text_length, copyright_length, title_length;
  long total_length ;

/* Declare the structure, initializing its pointers.  Default instrument 
is set to piano, channel has already been determined. */
  struct D_V detail_vars = { text, copyright, title, instrument, channel, patch} ;

/* Get more info from the user, and take it out of the structure */
  query_details(&detail_vars) ;
  patch = detail_vars.patch ; 

/* Add up all the lengths of the events */
  total_length = 0 ;

/* add the text event, if any */
  text_length = strlen(text) ;
  if (text_length > 0) { 
    total_length += 3 ;  /* delta time of zero, meta, type */
    total_length += write_var_bytes(text_length, var_bytes) ;  
    total_length += text_length ;
    use_text = TRUE ;
  }
  else {
    use_text = FALSE ;
  }

/* add the copyright, if any */
  copyright_length = strlen(copyright) ;
  if (copyright_length > 0) {
    total_length += 3 ;
    total_length += write_var_bytes(copyright_length, var_bytes) ;
    total_length += copyright_length ;
    use_copyright = TRUE ;
  }
  else {
    use_copyright = FALSE ;
  }
  
/* add the title event */
  total_length += 3 ;
  title_length = strlen(title) ;
  total_length += write_var_bytes(title_length, var_bytes) ;
  total_length += title_length ;

/* add an instrument event, unless this is a drum track (ie channel == 9) */
  if (channel != 9) {
    total_length += 3 ;
    total_length += write_var_bytes(strlen(instrument), var_bytes) ;
    total_length += strlen(instrument) ;
  }

  /* account for the time signature event */
  total_length += 8 ; 

 /* account for the key signature event */
  total_length += 6 ; 
  
 /* account for the initial tempo event */
  total_length += 7 ;

/* account for the initial (and only) program change 
             (ie - voice select) event */
  total_length += 3 ;

/* add the channel events recorded */
  total_length += count_bytes ;

/* account for the end of track event */
  total_length += 4 ;

/* Open the output file, taking the filename from the title */
  strcpy(filename, title) ;
  strcat(filename, ".mid") ;
/* If the title has weird characters, make a default name */
  for (i = 0 ; title[i] ; ++i) {
    if (isalnum(title[i]) == 0) {
      sprintf(take_str, "%i", take) ;
      strcpy(filename, "tune") ;
      strcat(filename, take_str) ;
      strcat(filename, ".mid") ;
      break ;
    }
  }

/* Change the default file permissions of the output, to resolve the 
annoyance of the output being a read only file.  A reasonable choice for 
the permissions seems readable for all, writable for owner and group, 
and executable for none. That is the numeric value 0664.  Running with 
suid makes things more complex!  So, instead let's live dangerously and 
give read and write permission to all. To try to get this to work, the 
permission macros defined in <sys/stat.h> shall be used.
*/
#ifdef LINUX_VERSION
  int orig_mask ;
  int rqd_mask = (S_IXUSR | S_IXGRP | S_IXOTH) ;
  orig_mask = umask(rqd_mask) ;
#endif

  if ((output = fopen(filename, "w+b")) == NULL) {
    fprintf(stderr, "Couldn't open file:  %s.\n", filename) ;
/* perhaps not proper use of the EXIT_FAILURE macro, but main() checks 
for the same constant */
    return EXIT_FAILURE ; 
  }

/* Write the file header.  A bunch of arbitrary constants. */
  int fheader[] = {0X4D, 0X54, 0X68, 0X64, 0X00, 0X00, 0X00, 0X06,
                  0X00, 0X00, 0X00, 0X01, 
                  TICKS_PER_BEAT / 0X100, TICKS_PER_BEAT % 0X100 } ;

  for (i = 0 ; i < 14 ; ++i)
    putc(fheader[i], output) ;

/* Write the track header */
  int theader[] = {0X4D, 0X54, 0X72, 0X6B} ;
  for (i = 0 ; i < 4 ; ++i)
    putc(theader[i], output) ;

/* Size in four bytes, to make sure they're big-endian, do them one at a time */
  for (i = 3 ; i >= 0 ; --i) {
    c = (total_length & (0XFF << (i * 8))) >> (i * 8) ;
    putc(c, output) ;
  }

/* Write text event */
  if (use_text) {
    putc(0X00, output) ;    /* delta-time of zero */
    putc(0XFF, output) ;     /* meta-event */
    putc(0X01, output) ;     /* type: text  */

/* bytes_rqd to exceed one is not anticipated, but let's be safe */
    bytes_rqd = write_var_bytes(text_length, var_bytes) ;
    for (i = 0 ; i < bytes_rqd ; ++i)
      putc(var_bytes[i], output) ;
    fputs(text, output) ;
  }
  
/* Write copyright event */
  if (use_copyright) {
    putc(0X00, output) ;
    putc(0XFF, output) ; 
    putc(0X02, output) ;            /* type: copyright  */
    bytes_rqd = write_var_bytes(copyright_length, var_bytes) ;
    for (i = 0 ; i < bytes_rqd ; ++i)
      putc(var_bytes[i], output) ;
    fputs(copyright, output) ;
  }

/* Write title event */
  putc(0X00, output) ;
  putc(0XFF, output) ;
  putc(0X03, output) ; 
  bytes_rqd = write_var_bytes(title_length, var_bytes) ;        
  for (i = 0 ; i < bytes_rqd ; ++i)
    putc(var_bytes[i], output) ;
  fputs(title, output) ;

/* Write instrument event */
  if (channel != 9) {
    putc(0X00, output) ;
    putc(0XFF, output) ;
    putc(0X04, output) ;
    bytes_rqd = write_var_bytes(strlen(instrument), var_bytes) ;
    for (i = 0 ; i < bytes_rqd ; ++i)
      putc(var_bytes[i], output) ;
    fputs(instrument, output) ;
  }

/* Write time signature */
  putc(0X00, output) ;
  putc(0XFF, output) ;
  putc(0X58, output) ;
  putc(0X04, output) ;
  putc(time_sig, output) ;
  putc(0X02, output) ;
  putc(0X18, output) ;
  putc(0X08, output) ;

/* Write the key signature */
  putc(0X00, output) ;
  putc(0XFF, output) ;
  putc(0X59, output) ;
  putc(0X02, output) ;
  putc( ((signed char)key_sig), output) ;
  putc(0X00, output) ;                    /* always a major key */

/* Write an initial tempo event */
  putc(0X00, output) ;
  putc(0XFF, output) ;
  putc(0X51, output) ;
  putc(0X03, output) ;
/* Three bytes must be written one at a time */
  for ( i = 2 ; i >= 0 ; --i) {
    c = (tempo >> (i * 8)) & 0XFF ;
    putc(c, output) ;
  }

/* Write a program change event */
  putc(0X00, output) ;
  c = (0X0C << 4) | channel ;
  putc(c, output) ;
  putc(patch, output) ;

/* Now for the recorded events.  Loop, advancing the pointer */
  ptr_active = ptr_head ;
  do {

/* The delta time is the same routine for either event */
    i = 0 ;
    do {
      putc( ptr_active -> delta_time[i], output) ;
      ++i ;
    } while ( ptr_active -> delta_time[i-1] & 0X80 ) ;

/* Events off, on, aftertouch may be written with the same routine */
    if ( (ptr_active -> type == 0X8)   || 
         (ptr_active -> type == 0X9)   || 
         (ptr_active -> type == 0XA) )   
    {
      c = (ptr_active -> type << 4) | (ptr_active -> channel) ;
      putc(c, output) ;
      putc( ptr_active -> note, output) ;
      putc( ptr_active -> velocity, output) ;
    }

/* The 'Channel Aftertouch' event is slightly different */
    else if (ptr_active -> type == 0XD) {
      c = (ptr_active -> type << 4) | (ptr_active -> channel) ;
      putc(c, output) ;
      putc( ptr_active -> velocity, output) ;
    }

/*
 Write the tempo change event.  Recall that the tags in the struct
 have nothing to do with what the data represents in this case! 
*/ 
    else if ( ptr_active -> type == 0X51) {
      putc(0XFF, output) ;                      /* Meta */
      putc(ptr_active -> type, output) ;       /* Tempo */
      putc(0X03, output) ;                     /* bytes following */      
      putc(ptr_active -> channel, output) ;    /* 1st byte of tempo */
      putc(ptr_active -> note, output) ;       /* 2nd byte of tempo */
      putc(ptr_active -> velocity, output) ;   /* 3rd byte of tempo */
    }
    else {           /* something must be horribly wrong! */
      exit(EXIT_FAILURE) ;         /* invalid event type */
    }

    ptr_active = ptr_active -> ptr_next ;
  } while (ptr_active != NULL) ;   

/* Write the end of track event, use an arbitrary delta time */
  putc(0X7F, output) ;
  putc(0XFF, output) ;
  putc(0X2F, output) ;
  putc(0X00, output) ;

/* Done writing the file.  Restore the old umask value */
  fclose(output) ;

#ifdef LINUX_VERSION
  umask(orig_mask) ;
#endif

  return EXIT_SUCCESS ;
} /* End Function write_file() */

/**********************************************************************/

/* shut off the speaker and eject from the program on an interrupt */
void handle_signal(int signum)
{
  if (signum == SIGINT) {

/* All the cleanup is registered with atexit(), so not much to do! */
    fprintf(stderr, "Exit by user interrupt.\n") ;
    exit(EXIT_FAILURE) ;  /* Because a good Midi file was not written */
  } 
} /* End Function handle_signal */

/**********************************************************************/

/* create and connect a new node and advance the ptr_active to it */
DBL_NODE *add_node(DBL_NODE *ptr_active)
{
  DBL_NODE *ptr_new ;
  ptr_new = (DBL_NODE *)malloc(sizeof(DBL_NODE)) ;
  if (ptr_new == NULL) {
    fprintf(stderr, "Memory allocation error\n") ;
    exit(EXIT_FAILURE) ;
  }

/* If the list has been cleared, start a new linked list */
  if (ptr_head == NULL) {
    ptr_head = ptr_new ;
    ptr_new -> ptr_prev = NULL ;
    ptr_new -> ptr_next = NULL ;
  }
  else {
    ptr_active -> ptr_next = ptr_new ;
    ptr_new -> ptr_prev = ptr_active ;
    ptr_new -> ptr_next = NULL ;
  }

  return ptr_new ;
}

/**********************************************************************/
/* Wipe out the whole linked list, clearing the memory */
void erase_list()
{
  DBL_NODE *ptr_active ;

/* in case nothing was recorded */
  if (ptr_head == NULL)
    return ;

  ptr_active = ptr_head ;
  while ( ptr_active -> ptr_next != NULL) {
    ptr_active = ptr_active -> ptr_next ;
    free( ptr_active -> ptr_prev ) ;
  }
  free(ptr_active) ;
  ptr_head = NULL ;

}   /* End Function erase_list */

/**********************************************************************/
