/* imidiacy - Record a MIDI file from the computer keyboard 
version 2007.6.14 Copyright (C) 2006-2007 Andrew Allinger 

***********************************************************************
This program is free software, you can redistribute it and/or modify it 
under the terms of the GNU General Public License, as published by the 
Free Software Foundation, version 2.

This program is distributed in hope that it will be useful but WITHOUT 
ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not write the Free Software Foundation, Inc. 59 
Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

This program produces standard MIDI files based on presses of the PC
keyboard.  The events Note Off, Note On, Aftertouch, and tempo change
are supported.  Output is to the monitor and soundcard or PC speaker.
 
For the most recent released version, visit:
imidiacy.sourceforge.net

Or, contact the author at:
andy_a@users.sourceforge.net

***********************************************************************/

/* Standard C includes */
#include <stdio.h>
#include <math.h>

/* Weird Linux stuff */
/* required for ioctl() */
#include <sys/ioctl.h>     

/* defines KIOCSOUND, which represents the PC speaker */
#include <sys/kd.h>

/* required for the low-level I/O functions open() and close() */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#define CLOCK_TICK_RATE 1193180  /* speed of the old PC clock */

/* the file descriptor should be kept track of by a global variable, so 
as not to have to pass it back to the main program */
int console_fd = -1 ;    /* initially not ready */

/**********************************************************************/

/* Turn on the PC speaker */
void play_note(int note, int velocity) 
{
  float frequency ;
  frequency = 440.0 * pow(2.0, (note - 69) / 12.0 ) ;

/* No way (known the the programmer) to adjust volume of the speaker */
  (void) velocity ;

  if ( ioctl(console_fd, KIOCSOUND, (int)(CLOCK_TICK_RATE / frequency)) == -1) {
    perror("Unable to turn on speaker") ;
  }
} 

/*************************************************************************/ 
/* Dummy function to match the one in audio_fns.c No good way available 
to play drum sounds on the speaker, within the capacity of the present 
programmer */

void play_drum(int note, int velocity) 
{
  play_note(note, velocity) ;
}

/******************************************************************************/

/* Turn off the PC speaker */
void hush_note(int note) 
{  
/* Don't need the note number here */
  (void) note ;

/* If the speaker was never opened, don't turn it off */
  if (console_fd < 0)  return ;

  if ( ioctl(console_fd, KIOCSOUND, 0) == -1) { 
    perror("Unable to turn off speaker") ;
  }
} 

/******************************************************************************/
/* Needed here to match the function in audio_fns.c , this does nothing */

void adjust_volume(int note, int velocity) 
{
  (void) note ;
  (void) velocity ;
  return ;
}
  
/******************************************************************************/
/* connect file descriptor to the console */ 
int init_sound() 
{
  if ( (console_fd = open("/dev/console", O_WRONLY)) < 0 ) {
    perror("Couldn't get access the speaker") ;
  }

/* Return a value here because this is where an error is most likely to 
occur.  This value will be called something else in the main program, 
and will not be passed to the functions in the present source module. */
  return console_fd ;
}

/******************************************************************************/

/* Shut down the speaker */
void end_sound() 
{ 
/* If the speaker was never opened, don't turn it off */
  if (console_fd < 0)  return ;

  if ( close(console_fd) == -1) {
    perror("Couldn't release the speaker") ;
  }
}

/******************************************************************************/

