/* imidiacy - Record a MIDI file from the computer keyboard 
version 2007.9.06 Copyright (C) 2006-2007 Andrew Allinger 

***********************************************************************
This program is free software, you can redistribute it and/or modify it 
under the terms of the GNU General Public License, as published by the 
Free Software Foundation, version 2.

This program is distributed in hope that it will be useful but WITHOUT 
ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not write the Free Software Foundation, Inc. 59 
Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

This program produces standard MIDI files based on presses of the PC
keyboard.  The events Note Off, Note On, Aftertouch, and tempo change
are supported.  Output is to the monitor and soundcard or PC speaker.
 
For the most recent released version, visit:
imidiacy.sourceforge.net

Or, contact the author at:
andy_a@users.sourceforge.net

***********************************************************************/

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <ncurses.h>

/* to simplify, Linux-specifics are also in this module */
#include <unistd.h>
#include <sys/time.h>
#include "midi_types.h"   

/* The cursor y position.  This variable should be used in the present 
module only. */
int pos = 0 ; 

/**********************************************************************/
void legal(char *string) 
{
  char a = '\n' ;
  addstr(string) ;
  addch(a) ;
  addstr("This program is free software, according to the terms of the GPL.\n") ;
  addstr("It has ABSOLUTELY NO WARRANTY.  \n") ;

//  sleep(3) ; 
  press_any_key() ;
}

/**********************************************************************/
/* conserve processor time */
void delay(unsigned int milliseconds) 
{
/* Structures used by nanosleep() */
  struct timespec treq, trem ;
  treq.tv_sec = milliseconds / 1000 ;
  treq.tv_nsec = (milliseconds % 1000) * 1000000L ;
  nanosleep(&treq, &trem) ;
}

/**********************************************************************/
/* Just a placeholder to match the function in gfx_fns.c */
void prepare_timing()
{
}  
/**********************************************************************/

/* Show some visual feedback about the note currently sounding. */

void add_note(int note, int velocity, int key_sig, int channel)
{
  (void) channel ;
  int y, x, ymax, xmax ;
  bool black_key = FALSE ;     /* an initial value to avoid nettlesome warnings */
  char n = 'Z' ; /* an initial value which should never show up in the output! */

  switch (note % 12) {
    case 0 :
      black_key = FALSE ;
      n = 'C' ;
      break ;
    case 1 :
      black_key = TRUE ;
      n = (key_sig < 0) ? 'D' : 'C' ;
      break ;
    case 2 :
      black_key = FALSE ;
      n = 'D' ;
      break ;
    case 3 :
      black_key = TRUE ;
      n = (key_sig < 0) ? 'E' : 'D' ;
      break ;
    case 4 :
      black_key = FALSE ;
      n = 'E' ;
      break ;
    case 5 :
      black_key = FALSE ;
      n = 'F' ;
      break ;
    case 6 :
      black_key = TRUE ;
      n = (key_sig < 0) ? 'G' : 'F' ;
      break ;
    case 7 :
      black_key = FALSE ;
      n = 'G' ;
      break ;
    case 8 :
      black_key = TRUE ;
      n = (key_sig < 0) ? 'A' : 'G' ;
      break ;
    case 9 :
      black_key = FALSE ;
      n = 'A' ;
      break ;
    case 10 :
      black_key = TRUE ;
      n = (key_sig < 0) ? 'B' : 'A' ;
      break ;
    case 11 :
      black_key = FALSE ;
      n = 'B' ;
      break ;
  }  /* End Switch */

/* Find what the size of the screen is */
  getmaxyx(stdscr, ymax, xmax) ;

y = (int)( (ymax - 1) - (float)(ymax - STATUS_LINES - 1) * ((float)(velocity) / 127.0) );
x = note - LOWEST_SCREEN_NOTE ;

/* if coordinates are out of bounds, quit the function */
if ( (y < 0) || (y >= ymax) || (x < 0) || ((black_key?(x+1):x) >= xmax) )
  return ;

move(y, x) ;
addch(n) ;

/* Put in the accidental sign if needed.  The cursor advances automatically. */

if (black_key) {
  addch( ((key_sig < 0) ? 'b' : '#') ) ;
}
return ;
}        /* End Function add_note */

/**********************************************************************/

/* Delete display of the sounding note from the screen */

void del_note(int note, int velocity, int channel)
{
  (void) channel ;
  int y, x, ymax, xmax ;
  bool black_key = FALSE ;   /* avoid annoying warnings */
  getmaxyx(stdscr, ymax, xmax) ;


switch (note % 12) {
  case 0 :
    black_key = FALSE ;
    break ;
  case 1 :
    black_key = TRUE ;
    break ;
  case 2 :
    black_key = FALSE ;
    break ;
  case 3 :
    black_key = TRUE ;
    break ;
  case 4 :
    black_key = FALSE ;
    break ;
  case 5 :
    black_key = FALSE ;
    break ;
  case 6 :
    black_key = TRUE ;
    break ;
  case 7 :
    black_key = FALSE ;
    break ;
  case 8 :
    black_key = TRUE ;
    break ;
  case 9 :
    black_key = FALSE ;
    break ;
  case 10 :
    black_key = TRUE ;
    break ;
  case 11 :
    black_key = FALSE ;
    break ;
}

y = (int)( (ymax - 1) - (float)(ymax - STATUS_LINES - 1) * ((float)(velocity) / 127.0) );
x = note - LOWEST_SCREEN_NOTE ;

/* if the coordinates are out of bounds, quit the function */
if ( (y < 0) || (y >= ymax) || (x < 0) || ((black_key?(x+1):x) >= xmax) )
  return ;

move(y, x) ;
delch() ;

/* if it's a black key, erase the accidental sign too */
if (black_key)
  delch() ;

return ;
}  /* End Function del_note */

/**********************************************************************/
/* adjust the position of the slider indicator for tempo or velocity */

void slider(int linenumber, int old_setting, int new_setting)
{
int left_chars, right_chars, x, range, min_setting, ymax, xmax ;
float quotient ;
  getmaxyx(stdscr, ymax, xmax) ;

if (linenumber == TEMPO_LINE) {
  left_chars = 8 ;   /* chars, disregarding null byte, in "Adagio <" */
  right_chars = 9 ;     /* characters in "> Allegro" */
  range = MAX_BPM - MIN_BPM ;
  min_setting = MIN_BPM ;
}
if (linenumber == VELOCITY_LINE) {
  left_chars =  7 ;  /* characters in "Piano <" */
  right_chars = 7 ;   /* characters in "< Forte"  */
  range = 127 - 0 ;
  min_setting = 0 ;
}

/* erase the old setting mark */
quotient = (float)(old_setting - min_setting) / (float)range ;
x = left_chars + (int)((float)(xmax - (left_chars + right_chars) -1) * quotient) ;
if ( (x >= 0) && (x < xmax) && (linenumber >= 0) && (linenumber < ymax) ) {
  move(linenumber, x) ;
  addch('-') ;
}

/* draw a new setting mark */
quotient = (float)(new_setting - min_setting) / (float)range ;
x = left_chars + (float)(xmax - (left_chars + right_chars) -1) * quotient ;
if ( (x >= 0) && (x < xmax) && (linenumber >= 0) && (linenumber < ymax) ) {
  move(linenumber, x) ;
  addch('|') ;
}

refresh() ;
}  /* End Function slider */
 
/**********************************************************************/
/* adjust the position of the slider indicator for legato */

void sliderf(int linenumber, float old_setting, float new_setting)
{
int left_chars, right_chars, x, xmax, ymax ;
float range, min_setting ;
  getmaxyx(stdscr, ymax, xmax) ;

if (linenumber == LEGATO_LINE) {
  left_chars = 10 ;        /* chars, disregarding null, in "Staccato <" */
  right_chars = 8 ;        /* characters in "> Legato" */
  range = 1.0 - 0.5 ;      /* the limits have been hard-coded */
  min_setting = 0.5 ;        
}

/* erase the old setting mark */
x = left_chars + (float)(xmax - (left_chars + right_chars) -1) * ((old_setting - min_setting) / range) ;
if ( (x > 0) && (x < xmax) && (linenumber > 0) && (linenumber < ymax) ) {
  move(linenumber, x) ;
  addch('-') ;
}

/* draw a new setting mark */
x = left_chars + (float)(xmax - (left_chars + right_chars) -1) * ((new_setting - min_setting) / range) ;
if ( (x > 0) && (x < xmax) && (linenumber > 0) && (linenumber < ymax) ) {
  move(linenumber, x) ;
  addch('|') ;
}

}  /* End Function sliderf */

/**********************************************************************/

/* 
Put a few informational lines up on the screen to initialize.  Nothing 
fancy, but it's better than having the screen go blank, which would 
likely disturb any user.
*/

void draw_display(int bpm, int velocity, float legato, char *version_string)
{
  int i, ymax, xmax ;
  getmaxyx(stdscr, ymax, xmax) ;

/* Start at the upper-left and print the program title line */
  clear() ;
  move(0, 0) ;
  addnstr(version_string, -1) ;

/* Start on the next line and print the instruction line */
  move(1, 0) ;
  addnstr("F1 or Backspace=Quit Up/Down=Vol Left/Right=Tempo \
Home/End=Legato PgUp/PgDn=f/p.", -1) ;


/* Print the Tempo line */
  if (move(TEMPO_LINE, 0) != ERR) {
    addnstr("Adagio <", -1) ;                   /*  Length w/o null:  8   */
    for (i =  8 ; i < (xmax - 9) ; ++i) {    
      addch('-') ;
    }
    addnstr("> Allegro", -1) ;                  /* Length w/o null:  9 */
  }
  else {
    addstr("Error:  Coordinates out of bounds.") ;
  }

/*  Print the velocity Line */
  if (move(VELOCITY_LINE, 0) != ERR) {
    addnstr("Piano <", -1) ;                 /* Length w/o null: 7 */
    for (i = 7 ; (i < (xmax - 7)) ; ++i) {     
      addch('-') ;
    }
    addnstr("> Forte", -1) ;                  /* Length w/o null:  7 */
  }
  else {
    addstr("Error:  Coordinates out of bounds.") ;
  }

/* Print the legato line */
  if (move(LEGATO_LINE, 0) != ERR) { 
    addnstr("Staccato <", -1) ;                 /* Length w/o null: 10 */
    for (i = 10 ; i < (xmax - 8) ; ++i) {      
      addch('-') ;
    }
    addnstr("> Legato", -1) ;                   /* Length w/o null: 8 */
  }
  else {
    addstr("Error:  Coordinates out of bounds.") ;
  }

/* Print a metronome line */
  if (move(METRONOME_LINE, 0) != ERR) {
    addnstr("Metronome:  ", -1) ;
  }
  else {
    addstr("Error:  Coordinates out of bounds.") ;
  }

/* Set the initial slider positions.  There is no "old" setting, but using
the new setting twice should work fine to call the slider functions */
  slider(TEMPO_LINE, bpm, bpm) ;
  slider(VELOCITY_LINE, velocity, velocity) ;
  sliderf(LEGATO_LINE, legato, legato) ;

  refresh() ;
  nodelay(stdscr, TRUE) ;
}  /* End Function draw_display */

/**********************************************************************/
/* query the user with a string, for another string */
static char *dialog(char *question, char *answer)
{
  int ymax, xmax ; 
  getmaxyx(stdscr, ymax, xmax) ;

  if (pos >= ymax-1) {
    clear() ; 
    pos = 0 ;
    move(pos, 0) ;
  }
  else {
    move(pos, 0) ;
  }
  addstr(question) ;
  ++pos ;
  getnstr(answer, STRING_SIZE - 1) ;
  ++pos ;

/* Append a null at the end of the string */
  answer[STRING_SIZE - 1] = '\0' ;

  return answer ;
}  /* End Function dialog */

/**********************************************************************/
int query_channel()
{

/* 
I can't be expected to keep these alternate versions of the program 
exactly synchronized, now can I?  How would any further development 
occur if I'm stuck trying to fix stuff that isn't broken, and add 
features nobody cares about?
*/
  return 0 ;
}

/**********************************************************************/

void query_settings(struct S_V *setting_vars)
{
  char answer[STRING_SIZE] ;

  clear() ;
  nodelay(stdscr, FALSE) ;
  echo() ;
  move(pos, 0) ;

  do {         /* Query for bpm */
    setting_vars->bpm = atoi(dialog("Please enter a beginning tempo in beats per minute.  ", answer)) ;
  } while ( (setting_vars->bpm > MAX_BPM) || (setting_vars->bpm < MIN_BPM) ) ;

  do {         /* query for velocity */
    setting_vars->velocity = atoi(dialog("Please enter a beginning dynamic level from 0 to 127.  ", answer)) ;
  } while ( (setting_vars->velocity > 127) || (setting_vars->velocity < 0) ) ;

  do {   /* query for legato */
    setting_vars->legato = atof(dialog("Beginning legato level as a decimal (0.5=staccato, 1.0=slurred)?  ", answer)) ;
  } while ( (setting_vars->legato > 1.0) || (setting_vars->legato < 0.5) ) ;

  do {    /* query for channel */
    setting_vars->channel = atoi(dialog("Please assign a MIDI channel to this track from 0 to 15.  ", answer)) ;
  } while ( (setting_vars->channel > 15) || (setting_vars->channel < 0) ) ;

  do {     /* Ask the key signature */
    setting_vars->key_sig=atoi(dialog("How many sharps in the key signature (flats count negative)?   ", answer)) ;
  } while ( (setting_vars->key_sig > 7) || (setting_vars->key_sig < -7) ) ;
 
  do {     /* Ask the time signature */
    setting_vars->time_sig = atoi(dialog("How many beats per measure?  ", answer)) ;
  } while ( (setting_vars->time_sig < 2) || (setting_vars->time_sig > 7) ) ;

/* Never use harmony in ncurses version */
  setting_vars->harmony = FALSE ;

}  /* End Function query_settings */

/**************************************************************************/

/* Ask user:  keep the take? */
int query_keep() 
{ 
  int ymax, xmax ;
  bool retval = TRUE ;
  unsigned char a ;
  getmaxyx(stdscr, ymax, xmax) ;

/* Clear the screen, and adjust modes */
  while(getch() != ERR) ;
  clear() ;
  refresh() ;
  pos = 0 ;
  nodelay(stdscr, FALSE) ;

  do {
    if (pos >= ymax-1) {
      clear() ;
      pos = 0 ;
      move(pos, 0 ) ;
    } 
    else {
      move(pos, 0) ;
    }
 
    addstr("Keep this take?  (Y/N) ") ;
    a = getch() ; 
    if ( (a == 'Y') || (a == 'y')) {
      retval = TRUE ;
    } 
    else if ( (a == 'N') || (a == 'n')) {
      retval = FALSE ;
    }
    else {
      a = '\0' ;
    }
    ++pos ; 
  } while (!a) ;

  return retval ;
}  /* End Function query_keep */

/**********************************************************************/

/* return 0 for quit, 1 for another take, and 2 to change settings */
int query_another() 
{
  int ymax, xmax ;
  unsigned char a ;
  int retval = 2 ;
  getmaxyx(stdscr, ymax, xmax) ;

  noecho() ;

  do {
    if (pos >= ymax-1) {
      clear() ;
      pos = 0 ;
      move(pos, 0 ) ;
    } 
    else {
      move(pos, 0) ;
    }

    addstr("Another take, change settings, or quit?  (T/S/Q) \n") ;
    a = getch() ;
    ++pos ;
    if ( (a == 'T') || (a == 't')) {
      retval = 1 ;
      return retval ;
    } 
    else if ( (a == 'S') || (a == 's')) {
      retval = 2 ;
      return retval ;
    }
    else if ( (a == 'Q') || (a == 'q') ) {
      retval = 0 ;
      return retval ;
    }
    else {
      a = '\0' ;
    }
  } while (!a) ;
  return retval ;
} /* End Function query_another */

/**********************************************************************/
/* tranlate a patch number to an English name for the instrument */

static char *tran_inst(int patch)
{

/* Names of the patches available to timidity.  Some of these voices are 
not available in freepats.  But it seems like a good idea to make them 
available anyway, so that this limitation lies outside of the program 
rather than inherent in it.  Names here follow freepats where available, 
Yamaha, elsewhere. */

  static char *strings[] = 

  { 
"Piano", "Brite Piano", "Electric G Piano", "Honky-tonk Piano", 
"Electric Piano 1", "Electric Piano 2", "Harpsichord", "Clavinet", 
"Celesta", "Glockenspiel", "Music Box", "Vibraphone", "Marimba", 
"Xylophone", "Tubular Bells", "Dulcimer", "Hammond Organ", "Percussive \
Organ", "Rock Organ", "Church Organ", "Reed Organ", "Accordion", 
"Harmonica", "Tango Accordion", "Nylon Guitar", "Steel Guitar", "Jazz \
Guitar", "Clean Electric Guitar", "Muted Electric Guitar", "Overdriven \
Guitar", "Distortion Guitar", "Guitar Harmonics", "Acoustic Bass", 
"Finger Bass", "Pick Bass", "Fretless Bass", "Slap Bass 1", "Slap Bass \
2", "Synth Bass 1", "Synth Bass 2", "Violin", "Viola", "Cello", 
"Contrabass", "Tremolo Strings", "Pizzicato Strings", "Harp", "Timpani", 
"String Ensemble", "String Ensemble 2", "Synth Strings 1", "Synth \
Strings 2", "Choir Aahs", "Voice Oohs", "Synth Voice", "Orchestra Hit", 
"Trumpet", "Trombone", "Tuba", "Muted Trumpet", "French Horn", "Brass \
Section", "Synth Brass 1", "Synth Brass 2", "Soprano Sax", "Alto Sax", 
"Tenor Sax", "Baritone Sax", "Oboe", "English Horn", "Bassoon", 
"Clarinet", "Piccolo", "Flute", "Recorder", "Panpipe", "Bottle Blow", 
"Shakuhachi", "Whistle", "Ocarina", "Square Wave", "Sawtooth Lead", 
"Calliope Lead", "Chiff Lead", "Charang Lead", "Voice Lead", "Fifth \
Lead", "Bass+Lead", "New Age", "Warm Pad", "Polysynth Pad", "Choir Pad", 
"Bowed Pad", "Metallic Pad", "Halo Pad", "Sweep Pad", "Rain", 
"Soundtrack", "Crystal", "Atmosphere", "Brightness", "Goblins-Unicorn", 
"Echo Voice", "Sci-Fi", "Sitar", "Banjo", "Shamisen", "Koto", "Kalimba", 
"Bagpipe", "Fiddle", "Shanai", "Tinkle Bell", "Agogo", "Steel Drums", 
"Wood Block", "Taiko Drum", "Melodic Drum", "Synth Drum", "Reverse \
Cymbal", "Fret Noise", "Breath Noise", "Seashore", "Bird Tweet", 
"Telephone Ring", "Helicopter", "Applause", "Gunshot"
  } ;

  return strings[patch] ;

} /* End Function *tran_inst() */

/**********************************************************************/

void query_details(struct D_V *detail_vars)
{
  unsigned char a ; 
  char answer[STRING_SIZE] ;
  int ymax, xmax ;
  getmaxyx(stdscr, ymax, xmax) ;

  echo() ;

/* Query user for title information */
  do {    
    dialog("Please enter a title for this track.\n", detail_vars->title) ;
  } while (strlen(detail_vars->title) == 0 ) ;

/* Query for copyright and text strings */
  dialog("Author and date?\n", detail_vars->copyright) ;
  dialog("Any additional comments?\n", detail_vars->text) ;

/* If this is a drum track */
  if (detail_vars->channel == 9) {
    detail_vars->patch = 0 ;
    detail_vars->instrument[0]  = '\0' ;
  }
/* If not a drum track, find out the instrument. */
  else {
    do {  /* Loop until user approves choice of an patch/instrument */   
      a = 'N' ;
      detail_vars->patch = atoi(dialog("Which voice to play, as a MIDI patch number 0-127? \n", answer)) ;

/* All these strings are of known size, so this ought to be safe */
      strcpy(detail_vars->instrument, tran_inst(detail_vars->patch) ) ;
      if ( (detail_vars->patch < 0) || (detail_vars->patch > 127) )  {
        pos += 2 ;
        continue ;
      }
/* This condition is obsolete */
      if (detail_vars->instrument[0] == '\0') {
        addstr("Sorry, this voice is unavailable.\n") ;
        pos += 2 ;
        continue ;
      }
      do {
        if (pos >= ymax-1) {
          clear() ;
          pos = 0 ;
          move(pos, 0 ) ;
        } 
        else {
          move(pos, 0) ;
        }

        addstr("You have selected voice ") ;
        addstr(detail_vars->instrument) ;
        addstr(" Okay?  (Y/N)") ;
        a = getch() ;
        pos = pos + 1 ;
      } while ( (a != 'y') && (a != 'Y') && (a != 'n') && (a != 'N') ) ;
    } while ( (a == 'N') || (a == 'n') ) ;  
  }  /* End If */
} /* End Function query_details */

/**********************************************************************/

void metronome(unsigned long present_time, int time_sig)
{
  int i, x ;

/* Erase the old mark.  An earlier version tried to locate the old mark, 
and precisely erase just what needed be removed.  That proved 
aggravating, and the present method is to inelegantly suck the whole 
mess off the screen */
  move(METRONOME_LINE, 12) ;
  for (i = 0 ; i <= time_sig ; ++i)
    delch() ;
 
/* x is the present beat of the measure.  Draw a new mark */
  x = ((present_time / TICKS_PER_BEAT) % time_sig) + 1 ;
  move(METRONOME_LINE, x + 12) ;
  addch(x + (int)('1') - 1 ) ;  
  refresh() ;

} /* End Function metronome() */

/**********************************************************************/ 
/* a dummy function, just to match the one in the Allegro version */
void end_timing() 
{
  ;
}

/**********************************************************************/ 
/* a dummy function, just to match the one in the Allegro version */
void scroll_roll()
{
  refresh() ;
}

/**********************************************************************/ 
/* This function takes out the curses calls from record(). */

int get_input(int channel) {
  (void) channel ;
  int keystroke ;
  keystroke = getch() ;

/* translate a keystroke to an instruction.  No tricks in this function. */
  switch (keystroke) {

/* First and most important, check that input exists. */
    case ERR : return NO_KEYSTROKE ;

/*  the spacebar may be used to perform a rest */
    case ' ' :    return REST ;

/* the alphanumerics denote various tones, with upper case symbols 
representing the bass register, three octaves deeper in pitch.  There's no 
adjustment for drum channels.  That allows drums 21-96 to be recorded, 
and, well,  you can't hear them anyway.*/
    case 'z' :    return 57 ;
    case 'x' :    return 58 ;
    case 'c' :    return 59 ;
    case 'v' :    return 60 ;
    case 'b' :    return 61 ;
    case 'n' :    return 62 ;
    case 'm' :    return 63 ;
    case ',' :    return 64 ;  
    case '.' :    return 65 ;
    case '/' :    return 66 ;
    case 'a' :    return 67 ; 
    case 's' :    return 68 ;
    case 'd' :    return 69 ;
    case 'f' :    return 70 ;
    case 'g' :    return 71 ;
    case 'h' :    return 72 ;
    case 'j' :    return 73 ;
    case 'k' :    return 74 ;
    case 'l' :    return 75 ;
    case ';' :    return 76 ;
    case 'q' :    return 77 ;
    case 'w' :    return 78 ;
    case 'e' :    return 79 ;
    case 'r' :    return 80 ;
    case 't' :    return 81 ;
    case 'y' :    return 82 ;
    case 'u' :    return 83 ;
    case 'i' :    return 84 ;
    case 'o' :    return 85 ;
    case 'p' :    return 86 ;
    case '1' :    return 87 ;
    case '2' :    return 88 ;
    case '3' :    return 89 ;
    case '4' :    return 90 ;
    case '5' :    return 91 ;
    case '6' :    return 92 ;
    case '7' :    return 93 ;
    case '8' :    return 94 ;
    case '9' :    return 95 ;
    case '0' :    return 96 ;
    case 'Z' :    return 21 ;
    case 'X' :    return 22 ;
    case 'C' :    return 23 ;
    case 'V' :    return 24 ;
    case 'B' :    return 25 ;
    case 'N' :    return 26 ;
    case 'M' :    return 27 ;
    case '<' :    return 28 ;
    case '>' :    return 29 ;
    case '?' :    return 30 ;
    case 'A' :    return 31 ;
    case 'S' :    return 32 ;
    case 'D' :    return 33 ;
    case 'F' :    return 34 ;
    case 'G' :    return 35 ;
    case 'H' :    return 36 ;
    case 'J' :    return 37 ;
    case 'K' :    return 38 ;
    case 'L' :    return 39 ;
    case ':' :    return 40 ;
    case 'Q' :    return 41 ;
    case 'W' :    return 42 ;
    case 'E' :    return 43 ;
    case 'R' :    return 44 ;
    case 'T' :    return 45 ;
    case 'Y' :    return 46 ;
    case 'U' :    return 47 ;
    case 'I' :    return 48 ;
    case 'O' :    return 49 ;
    case 'P' :    return 50 ;
    case '!' :    return 51 ;
    case '@' :    return 52 ;
    case '#' :    return 53 ;
    case '$' :    return 54 ;
    case '%' :    return 55 ;
    case '^' :    return 56 ; 
    case '&' :    return 57 ;
    case '*' :    return 58 ;
    case '(' :    return 59 ;
    case ')' :    return 60 ;
 
/* Special values of the control keys */
    case KEY_DOWN :      return DEC_VELOCITY ;
    case KEY_UP :        return INC_VELOCITY ;
    case KEY_LEFT :      return DEC_TEMPO ;
    case KEY_RIGHT :     return INC_TEMPO ;
    case KEY_NPAGE :     return PIANO ;
    case KEY_PPAGE :     return FORTE ;
    case KEY_HOME :      return STACCATO ;
    case KEY_END :       return LEGATO ;
    case KEY_BACKSPACE : return QUIT ; 

/* May help to have another way to exit the program, as KEY_BACKSPACE 
doesn't seem to be detected sometimes, for example in xterm. */
    case KEY_ENTER :     return QUIT ;
    case KEY_F(1) :      return QUIT ;

/* Make sure that stray keystrokes don't have a positive value */
    default :    return KEY_IGNORED ;
  }    /* End Switch */

}    /* End Function get_input */

/**********************************************************************/
/* Another code extraction:  get the time, now measured in milliseconds */
unsigned long get_time()
{
  unsigned long msec_now ;
  time_t       t_now ;        /* seconds at present */
  suseconds_t ut_now ;       /* microseconds at present */        

/* structures used by gettimeofday */
  struct timeval tval ;
  struct timezone tzone ;

  while (gettimeofday(&tval, &tzone) ) ; /* loop until success */
  t_now = tval.tv_sec ;
  ut_now = tval.tv_usec ;

  msec_now = (unsigned long)(t_now * 1000 + ut_now / 1000) ;
  return msec_now ;
} /* End Function get_time() */

/**********************************************************************/

/* A tiny function called once at the beginning of main() */
void ready_screen() 
{
  initscr() ;
  cbreak() ;
  keypad(stdscr, TRUE) ;
  nonl() ;
}

/**********************************************************************/
/* Brief, but necessary functions */
void cleanup_screen() 
{
  endwin() ;
}

/**********************************************************************/

/* Press Any Key To Continue */

void press_any_key()
{
  noecho() ;
  addstr("Press any key to continue.  \n") ;
  (void) getch() ;
}

/**********************************************************************/
