/* imidiacy - Record a MIDI file from the computer keyboard 
version 2006.12.17 Copyright (C) 2006 Andrew Allinger 

***********************************************************************
This program is free software, you can redistribute it and/or modify it 
under the terms of the GNU General Public License, as published by the 
Free Software Foundation, version 2.

This program is distributed in hope that it will be useful but WITHOUT 
ANY WARRANTY, not even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not write the Free Software Foundation, Inc. 59 
Temple Place - Suite 330, Boston, MA, 02111-1307, USA
***********************************************************************

This program records keystrokes and produces MIDI files. Supported are 
Note Off, Note On, Aftertouch, and tempo changes.  Output is to the 
monitor and PC speaker.  Further development is underway.
 
For the most recent released version, visit:
imidiacy.sourceforge.net

Or, contact the author at:
andy_a@users.sourceforge.net

***********************************************************************/


/***********************************************************************

To compile this utility, issue:

user@host$ cd imidiacy-source
user@host$ gcc -O3 -o <executable_path/vb_check> bitwise.c utils/vb_check.c
user@host$ ln -s <executable_path/vb_check> ulong2vb
user@host$ ln -s <executable_path/vb_check> vb2ulong 

USAGE
Invoke the program with the appropriate link.  For example:

user@host$ vb2ulong AB 06

Entering a sequence of up to four bytes in HEXADECIMAL FORMAT.  Likewise,

user@host$ ulong2vb 79309

Will give the variable byte sequence.
**************************************************************************/


#include <stdio.h> 
#include <string.h>

/* Have to declare the subroutines */
unsigned long read_var_bytes(unsigned char *var_bytes) ;
int write_var_bytes(unsigned long number, unsigned char *var_bytes) ;

int main(int argc, const char **argv)
{
  char *command_name = 0 ;
  unsigned char var_bytes[4] ;
  unsigned int translated[4] ;
  unsigned long number ;
  int i, count_bytes ;

/* Remove any leading pathname from the command */
  if ( (command_name = strrchr(argv[0], '/')) != 0 )
    ++command_name ;
  else
    command_name = argv[0] ;

/* Bail out for too much input */
  if (argc > 5) 
  {
    fprintf(stderr, "Input sequence too long\n") ;
    return 2 ;
  }

/* Use the read_var_bytes function */
  if ( strcmp(command_name, "vb2ulong") == 0 ) 
  {

/* The index will be one greater for the arguments than the array */
    for (i = 0 ; i < (argc-1) ; ++i)
    {
      sscanf( argv[i+1], "%X", (translated + i) ) ;
      var_bytes[i] = (unsigned char) translated[i] ;
    }
    number = read_var_bytes(var_bytes) ;
    
    printf("The hexadecimal sequence:") ;
    for (i = 0 ; i < (argc-1) ; ++i) 
      printf("  %X ", translated[i]) ;
    printf("is interpreted as %lu \n", number) ;
  }

/* Use the write_var_bytes function */
  else if ( strcmp(command_name, "ulong2vb") == 0)
  {
    if (argc == 1) 
    {
       fprintf(stderr, "Please supply a number to translate. \n") ;
       return 3 ;
    }
    sscanf( argv[1], "%lu", &number) ;
    count_bytes = write_var_bytes(number, var_bytes) ;

    printf("The number %lu is interpreted:", number) ;
    for (i = 0 ; i < count_bytes ; ++i) 
      printf(" %X ", var_bytes[i]) ;
    printf("\n") ;
  }
  
  else 
  {
    fprintf(stderr, "Couldn't recognize command name.\n") ;
    return 1 ;
  }

  return 0 ;
} /* End Program */
